from dataclasses import (
    dataclass,
)
from deprecated import (
    deprecated,
)
from typing import (
    Generic,
    Optional,
    Type,
    TypeVar,
    Union,
)

_L = TypeVar("_L")
_R = TypeVar("_R")


@deprecated("use `UnionFactory` instead")  # type: ignore[misc]
def inr(val: _R, _left: Optional[Type[_L]] = None) -> Union[_L, _R]:
    return val


@deprecated("use `UnionFactory` instead")  # type: ignore[misc]
def inl(val: _L, _right: Optional[Type[_L]] = None) -> Union[_L, _R]:
    return val


@dataclass(frozen=True)
class UnionFactory(Generic[_L, _R]):
    def inl(self, value: _L) -> Union[_L, _R]:
        return value

    def inr(self, value: _R) -> Union[_L, _R]:
        return value



from fa_purity import (
    _iter_factory,
)
from fa_purity.cmd import (
    Cmd,
    unsafe_unwrap,
)
from fa_purity.cmd.core import (
    CmdUnwrapper,
    new_cmd,
)
from fa_purity.maybe import (
    Maybe,
)
from fa_purity.pure_iter.core import (
    PureIter,
)
from fa_purity.stream.core import (
    _Stream,
    Stream,
)
from typing import (
    Iterable,
    Optional,
    TypeVar,
    Union,
)

_T = TypeVar("_T")


def _chain_1(
    unchained: Stream[PureIter[_T]],
) -> Stream[_T]:
    draft = _Stream(unchained.unsafe_to_iter().map(_iter_factory.chain))
    return Stream(draft)


def _chain_2(
    unchained: PureIter[Stream[_T]],
) -> Stream[_T]:
    def _action(act: CmdUnwrapper) -> Iterable[_T]:
        iters = (act.unwrap(i.unsafe_to_iter()) for i in unchained)
        return _iter_factory.chain(iters)

    draft = _Stream(new_cmd(_action))
    return Stream(draft)


def chain(
    unchained: Union[Stream[PureIter[_T]], PureIter[Stream[_T]]],
) -> Stream[_T]:
    if isinstance(unchained, Stream):
        return _chain_1(unchained)
    return _chain_2(unchained)


def squash(stm: Stream[Cmd[_T]]) -> Stream[_T]:
    draft = _Stream(stm.unsafe_to_iter().map(_iter_factory.squash))
    return Stream(draft)


def consume(stm: Stream[Cmd[None]]) -> Cmd[None]:
    return Cmd.from_cmd(
        lambda: _iter_factory.deque(
            iter(unsafe_unwrap(a) for a in unsafe_unwrap(stm.unsafe_to_iter()))
        )
    )


def filter_opt(stm: Stream[Optional[_T]]) -> Stream[_T]:
    draft = _Stream(stm.unsafe_to_iter().map(_iter_factory.filter_none))
    return Stream(draft)


def filter_maybe(stm: Stream[Maybe[_T]]) -> Stream[_T]:
    return filter_opt(stm.map(lambda x: x.value_or(None)))


def until_none(stm: Stream[Optional[_T]]) -> Stream[_T]:
    draft = _Stream(stm.unsafe_to_iter().map(_iter_factory.until_none))
    return Stream(draft)


def until_empty(stm: Stream[Maybe[_T]]) -> Stream[_T]:
    return until_none(stm.map(lambda m: m.value_or(None)))



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
from typing import (
    Callable,
    Generic,
    NoReturn,
    Optional,
    Type,
    TypeVar,
    Union,
)

_S = TypeVar("_S")
_F = TypeVar("_F")
_T = TypeVar("_T")


@dataclass(frozen=True)
class UnwrapError(Exception, Generic[_S, _F]):
    container: "Result[_S, _F]"


@dataclass(frozen=True)
class _Success(Generic[_T]):
    value: _T


@dataclass(frozen=True)
class _Failure(Generic[_T]):
    value: _T


@dataclass(frozen=True)
class Result(Generic[_S, _F]):
    _value: Union[_Success[_S], _Failure[_F]]

    @staticmethod
    def success(val: _S, _type: Optional[Type[_F]] = None) -> Result[_S, _F]:
        return Result(_Success(val))

    @staticmethod
    def failure(val: _F, _type: Optional[Type[_S]] = None) -> Result[_S, _F]:
        return Result(_Failure(val))

    def map(self, function: Callable[[_S], _T]) -> Result[_T, _F]:
        if isinstance(self._value, _Success):
            return Result(_Success(function(self._value.value)))
        return Result(self._value)

    def alt(self, function: Callable[[_F], _T]) -> Result[_S, _T]:
        if isinstance(self._value, _Failure):
            return Result(_Failure(function(self._value.value)))
        return Result(self._value)

    def bind(self, function: Callable[[_S], Result[_T, _F]]) -> Result[_T, _F]:
        if isinstance(self._value, _Success):
            return function(self._value.value)
        return Result(self._value)

    def lash(self, function: Callable[[_F], Result[_S, _T]]) -> Result[_S, _T]:
        if isinstance(self._value, _Failure):
            return function(self._value.value)
        return Result(self._value)

    def swap(self) -> Result[_F, _S]:
        if isinstance(self._value, _Failure):
            return Result(_Success(self._value.value))
        return Result(_Failure(self._value.value))

    def apply(self, wrapped: Result[Callable[[_S], _T], _F]) -> Result[_T, _F]:
        return wrapped.bind(lambda f: self.map(f))

    def value_or(self, default: _T) -> Union[_S, _T]:
        if isinstance(self._value, _Success):
            return self._value.value
        return default

    def or_else_call(self, function: Callable[[], _T]) -> Union[_S, _T]:
        # lazy version of `value_or`
        if isinstance(self._value, _Success):
            return self._value.value
        return function()

    def to_union(self) -> Union[_S, _F]:
        return self._value.value

    def unwrap(self) -> Union[_S, NoReturn]:
        if isinstance(self._value, _Success):
            return self._value.value
        raise UnwrapError(self)

    def unwrap_fail(self) -> Union[_F, NoReturn]:
        if isinstance(self._value, _Failure):
            return self._value.value
        raise UnwrapError(self)


ResultE = Result[_T, Exception]  # type: ignore[misc]



from fa_purity.json.factory import (
    from_any,
)
from fa_purity.json.transform import (
    dumps,
)

test_data = from_any({"foo": {"nested": ["hi", 99]}}).unwrap()


def test_dumps() -> None:
    assert dumps(test_data).replace(
        " ", ""
    ) == '{"foo": {"nested": ["hi", 99]} }'.replace(" ", "")



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
from fa_purity import (
    _iter_factory,
)
from fa_purity.cmd import (
    Cmd,
)
from fa_purity.cmd.core import (
    CmdUnwrapper,
    new_cmd,
)
from fa_purity.frozen import (
    FrozenList,
)
from fa_purity.maybe import (
    Maybe,
)
import functools
from typing import (
    Callable,
    Generic,
    Iterable,
    TypeVar,
)

_T = TypeVar("_T")
_R = TypeVar("_R")


@dataclass(frozen=True)
class _Stream(
    Generic[_T],
):
    _new_iter: Cmd[Iterable[_T]]


class Stream(_Stream[_T]):
    def __init__(self, obj: _Stream[_T]):
        super().__init__(obj._new_iter)

    def map(self, function: Callable[[_T], _R]) -> Stream[_R]:
        draft: _Stream[_R] = _Stream(
            self._new_iter.map(lambda i: iter(map(function, i)))
        )
        return Stream(draft)

    def reduce(self, function: Callable[[_R, _T], _R], init: _R) -> Cmd[_R]:
        return self._new_iter.map(
            lambda i: functools.reduce(function, i, init)
        )

    def bind(self, function: Callable[[_T], Stream[_R]]) -> Stream[_R]:
        def _action(act: CmdUnwrapper) -> Iterable[_R]:
            items = act.unwrap(self.map(function)._new_iter)
            unchained = iter(act.unwrap(i._new_iter) for i in items)
            return _iter_factory.chain(unchained)

        draft: _Stream[_R] = _Stream(new_cmd(_action))
        return Stream(draft)

    def filter(self, function: Callable[[_T], bool]) -> Stream[_T]:
        draft = _Stream(
            self._new_iter.map(lambda i: iter(filter(function, i)))
        )
        return Stream(draft)

    def find_first(self, criteria: Callable[[_T], bool]) -> Cmd[Maybe[_T]]:
        return self._new_iter.map(
            lambda i: _iter_factory.find_first(criteria, i)
        ).map(Maybe.from_optional)

    def chunked(self, size: int) -> Stream[FrozenList[_T]]:
        draft = _Stream(
            self._new_iter.map(lambda i: _iter_factory.chunked(i, size))
        )
        return Stream(draft)

    def transform(self, function: Callable[[Stream[_T]], _R]) -> _R:
        return function(self)

    def to_list(self) -> Cmd[FrozenList[_T]]:
        return self._new_iter.map(tuple)

    def unsafe_to_iter(self) -> Cmd[Iterable[_T]]:
        # if possible iterables should not be used directly
        return self._new_iter



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
from fa_purity.frozen import (
    FrozenDict,
    FrozenList,
)
from fa_purity.json.primitive.core import (
    Primitive,
)
from typing import (
    Union,
)

UnfoldedJVal = Union[
    FrozenDict[str, "JsonValue"], FrozenList["JsonValue"], Primitive
]


@dataclass(frozen=True)
class JsonValue:
    _value: UnfoldedJVal

    def unfold(
        self,
    ) -> UnfoldedJVal:
        return self._value
